import unittest
import tempfile

import palletjack as pj
import pyarrow.parquet as pq
import numpy as np
import pyarrow as pa
import os
import itertools as it

n_row_groups = 5
n_columns = 7
chunk_size = 1 # A row group per
current_dir = os.path.dirname(os.path.realpath(__file__))

def get_table():
    # Generate a random 2D array of floats using NumPy
    # Each column in the array represents a column in the final table
    data = np.random.rand(n_row_groups, n_columns)

    # Convert the NumPy array to a list of PyArrow Arrays, one for each column
    pa_arrays = [pa.array(data[:, i]) for i in range(n_columns)]

    # Optionally, create column names
    column_names = [f'column_{i}' for i in range(n_columns)]

    # Create a PyArrow Table from the Arrays
    return pa.Table.from_arrays(pa_arrays, names=column_names)

class TestPalletJack(unittest.TestCase):
      
    def test_read_metadata_columns_rows(self):

        def validate_reading(parquet_path, index_path, row_groups, columns):
            # Reading using the original metadata
            pr = pq.ParquetReader()
            pr.open(parquet_path)
            org_data = pr.read_row_groups(row_groups)
            if len(columns) > 0:                
                org_data = org_data.select(columns)

            # Reading using the indexed metadata
            metadata = pj.read_metadata(index_path, row_groups=row_groups, columns=columns)
            pr = pq.ParquetReader()
            pr.open(parquet_path, metadata=metadata)

            pj_data = pr.read_row_groups(list(range(0, len(row_groups))))
            self.assertEqual(org_data, pj_data, f"row_groups={row_groups}, columns={columns}")

        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as tmpdirname:
            path = os.path.join(tmpdirname, "my.parquet")
            table = get_table()

            pq.write_table(table, path, row_group_size=chunk_size, use_dictionary=False, write_statistics=False, store_schema=False)
            index_path = path + '.index'
            pj.generate_metadata_index(path, index_path)

            all_columns = list(range(0, n_columns))
            all_row_groups = list(range(0, n_row_groups))
            for r in range(0, 4):
                for rp in it.permutations(all_row_groups, r):
                    for c in range(0, 4):
                        for cp in it.permutations(all_columns, c):
                            validate_reading(path, index_path, row_groups = rp, columns = cp)

    def test_read_row_group_metadata(self):
        
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as tmpdirname:
            path = os.path.join(tmpdirname, "my.parquet")
            table = get_table()

            pq.write_table(table, path, row_group_size=chunk_size, use_dictionary=False, write_statistics=False, store_schema=False)

            index_path = path + '.index'
            pj.generate_metadata_index(path, index_path)
            for r in range(0, n_row_groups):

                # Reading using the original metadata
                pr = pq.ParquetReader()
                pr.open(path)
                res_data_org = pr.read_row_groups([r], use_threads=False)

                # Reading using the indexed metadata
                metadata = pj.read_metadata(index_path, row_groups = [r])
                pr = pq.ParquetReader()
                pr.open(path, metadata=metadata)

                res_data_index = pr.read_row_groups([0], use_threads=False)
                self.assertEqual(res_data_org, res_data_index, f"Row={r}")

    def test_reading_row_groups(self):
        
        with tempfile.TemporaryDirectory(ignore_cleanup_errors=True) as tmpdirname:
            path = os.path.join(tmpdirname, "my.parquet")
            table = get_table()

            pq.write_table(table, path, row_group_size=chunk_size, use_dictionary=False, write_statistics=False, store_schema=False)

            index_path = path + '.index'
            pj.generate_metadata_index(path, index_path)

            # Reading using the original metadata
            pr = pq.ParquetReader()
            pr.open(path)
            res_data_org = pr.read_row_groups([2, 3, 4], use_threads=False)
            
            # Reading using the indexed metadata
            metadata = pj.read_metadata(index_path, row_groups = [2, 3, 4])
            pr = pq.ParquetReader()
            pr.open(path, metadata=metadata)

            res_data_index = pr.read_row_groups([0, 1, 2], use_threads=False)
            self.assertEqual(res_data_org, res_data_index)

    def test_reading_invalid_row_group(self):
        with tempfile.TemporaryDirectory() as tmpdirname:
            path = os.path.join(tmpdirname, "my.parquet")
            table = get_table()

            pq.write_table(table, path, row_group_size=chunk_size, use_dictionary=False, write_statistics=False, store_schema=False)

            index_path = path + '.index'
            pj.generate_metadata_index(path, index_path)
            
            with self.assertRaises(RuntimeError) as context:
                metadata = pj.read_metadata(index_path, [n_row_groups])

            self.assertTrue(f"Requested row_group={n_row_groups}, but only 0-{n_row_groups-1} are available!" in str(context.exception), context.exception)

    def test_reading_invalid_column(self):
            with tempfile.TemporaryDirectory() as tmpdirname:
                path = os.path.join(tmpdirname, "my.parquet")
                table = get_table()

                pq.write_table(table, path, row_group_size=chunk_size, use_dictionary=False, write_statistics=False, store_schema=False)

                index_path = path + '.index'
                pj.generate_metadata_index(path, index_path)
                
                with self.assertRaises(RuntimeError) as context:
                    metadata = pj.read_metadata(index_path, row_groups=[], columns=[n_columns])

                self.assertTrue(f"Requested column={n_columns}, but only 0-{n_columns-1} are available!" in str(context.exception), context.exception)

    def test_reading_invalid_index_file(self):
        with tempfile.TemporaryDirectory() as tmpdirname:
            path = os.path.join(tmpdirname, "my.parquet")
            table = get_table()

            pq.write_table(table, path, row_group_size=chunk_size, use_dictionary=False, write_statistics=False, store_schema=False)

            with self.assertRaises(Exception) as context:
                metadata = pj.read_metadata(path, [0])

            self.assertTrue(f"File '{path}' has unexpected format!" in str(context.exception), context.exception)

    def test_index_file_golden_master(self):
        with tempfile.TemporaryDirectory() as tmpdirname:
            index_path = os.path.join(tmpdirname, 'my.parquet.index')
            path = os.path.join(current_dir, 'data/sample.parquet')
            expected_index_path = os.path.join(current_dir, 'data/sample.parquet.index')
            pj.generate_metadata_index(path, index_path)

            # Read the expected output
            with open(expected_index_path, 'rb') as file:
                expected_output = file.read()

            # Read the expected output
            with open(index_path, 'rb') as file:
                actual_output = file.read()

            # Compare the actual output to the expected output
            self.assertEqual(actual_output, expected_output)

        pr = pq.ParquetReader()
        pr.open(path)
        metadata = pr.metadata
        row_groups = metadata.num_row_groups

        for r in range(0, row_groups):

            # Reading using the original metadata
            pr = pq.ParquetReader()
            pr.open(path)
            res_data_org = pr.read_row_groups([r], use_threads=False)

            # Reading using the indexed metadata
            metadata = pj.read_metadata(expected_index_path, row_groups = [r])
            pr = pq.ParquetReader()
            pr.open(path, metadata=metadata)

            res_data_index = pr.read_row_groups([0], use_threads=False)
            self.assertEqual(res_data_org, res_data_index, f"Row={r}")

if __name__ == '__main__':
    unittest.main()




from libcpp.vector cimport vector
from libcpp.memory cimport shared_ptr
from libc.stdint cimport uint32_t
from pyarrow._parquet cimport *

cdef extern from "palletjack.h":
    cdef void GenerateMetadataIndex(const char *parquet_path, const char *index_file_path) nogil except +
    cdef shared_ptr[CFileMetaData] ReadMetadata(const char *index_file_path, const vector[uint32_t] row_groups, const vector[uint32_t] columns) nogil except +



# distutils: include_dirs = .

import pyarrow as pa
import pyarrow.parquet as pq
from cython.cimports.palletjack import cpalletjack
from libcpp.string cimport string
from libcpp.memory cimport shared_ptr
from libcpp.vector cimport vector
from libc.stdint cimport uint32_t
from pyarrow._parquet cimport *

def generate_metadata_index(parquet_path, index_file_path):
    cpalletjack.GenerateMetadataIndex(parquet_path.encode('utf8'), index_file_path.encode('utf8'))

cpdef read_metadata(index_file_path, row_groups = [], columns = []):

    cdef shared_ptr[CFileMetaData] c_metadata
    cdef string encoded_path = index_file_path.encode('utf8')
    cdef vector[uint32_t] crow_groups = row_groups
    cdef vector[uint32_t] ccolumns = columns
    with nogil:
        c_metadata = cpalletjack.ReadMetadata(encoded_path.c_str(), crow_groups, ccolumns)

    cdef FileMetaData m = FileMetaData.__new__(FileMetaData)
    m.init(c_metadata)
    return m
